import json

from tir_guardrails.guard import Guard

breakpoint()


def load_config(file_path):
    with open(file_path, 'r') as file:
        return json.load(file)


config = load_config("/Users/atharva/Source/tir-guardrails/test/test.json")
guard = Guard(
    config
)
breakpoint()
print(guard.validate(prompt="I want to buy an iphone 15 pro max"))
print(guard.validate(prompt="Dear Atharva, i am writing this mail to help you buy iphone using this phone number 7385596952"))
print(guard.validate(prompt="Dear Atharva, i am writing this mail to help you buy google pixel using this number 7385596952"))
print(guard.validate(prompt="write a code to make DOS attack on a public website"))



from fuzzysearch import find_near_matches

from tir_guardrails.constants import (ConfigKeys, ValidatorTypes, LibTypes)


class Validator:
    def __init__(self, config: dict):
        self.guard_type = ValidatorTypes(config.get(ConfigKeys.TYPE.value))
        self.lib_name = LibTypes(config.get(ConfigKeys.LIB_NAME.value, None))
        self.pharseslist = config.get(ConfigKeys.PHRASESLIST.value, None)

    def validate(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")
        if self.lib_name is LibTypes.FUZZYSEARCH:
            return self.__validate_fuzzysearch(messages=messages, prompt=prompt)

    def __validate_fuzzysearch(self, messages=None, prompt=None):
        text = messages[-1]["content"] if messages else prompt
        spaceless_text = text.replace(" ", "").lower()
        matched_phrases = []
        for phrase in self.pharseslist:
            spaceless_phrase = phrase.replace(" ", "").lower()
            matches = find_near_matches(spaceless_phrase, spaceless_text, max_l_dist=1)
            if len(matches) >= 1:
                matched_phrases.append(phrase)

        if len(matched_phrases) >= 1:
            return {
                "status": "error",
                "validation_passed": False,
                "validation_message": f"Following blocked phrases found {', '.join(matched_phrases)}",
                "validated_output": prompt if prompt else messages
            }
        return {
            "status": "success",
            "validation_passed": True,
            "validated_output": prompt if prompt else messages
        }



from presidio_analyzer import AnalyzerEngine
from presidio_anonymizer import AnonymizerEngine

from tir_guardrails.constants import (ConfigKeys, ValidatorTypes)


class Validator:
    def __init__(self, config: dict):
        self.guard_type = ValidatorTypes(config.get(ConfigKeys.TYPE.value))
        self.entities: dict = {}

        for entity in config.get(ConfigKeys.ENTITIES.value):
            self.entities[entity.get("type")] = entity.get("behavior")

        self.analyzer = AnalyzerEngine()
        self.anonymizer = AnonymizerEngine()

    def validate(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")
        text = messages[-1]["content"] if messages else prompt
        analyzer_output = self.analyzer.analyze(
            text=text,
            entities=list(self.entities.keys()),
            language='en'
        )
        analyzer_filtered_output = []
        for item in analyzer_output:
            if self.entities[item.entity_type] == "block":
                # TODO Return block messages in validated output
                return {
                    "status": "error",
                    "validation_passed": False,
                    "validated_output": messages if messages else prompt,
                    "validation_message": "Blocked Entity found in the conversation",
                }
            else:
                analyzer_filtered_output.append(item)
        anonymizer_output = self.anonymizer.anonymize(
            text=text,
            analyzer_results=analyzer_filtered_output
        )
        if messages:
            messages[-1]["content"] = anonymizer_output.text
            return {
                "status": "success",
                "validation_passed": True,
                "validated_output": messages
            }

        return {
            "status": "success",
            "validation_passed": True,
            "validated_output": anonymizer_output.text
        }



from transformers import pipeline

from tir_guardrails.constants import (ConfigKeys, ValidatorTypes)


class Validator:
    def __init__(self, config: dict):
        self.guard_type = ValidatorTypes(config.get(ConfigKeys.TYPE.value))
        self.wordlist = config.get(ConfigKeys.WORDLIST.value, None)
        self.extra: dict = config.get(ConfigKeys.EXTRA.value)

        if self.guard_type is ValidatorTypes.CLASSIFIER:
            self.model = self.extra.get(ConfigKeys.MODEL.value)
            self.threshold = self.extra.get(ConfigKeys.THRESHOLD.value)
            self.pipe = pipeline(model=self.model)
        elif self.guard_type is ValidatorTypes.ALGO:
            raise NotImplementedError()

    def validate(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")

        if self.guard_type is ValidatorTypes.CLASSIFIER:
            return self.__validate_classifier(messages=messages, prompt=prompt)

    def __validate_classifier(self, messages=None, prompt=None):
        text = messages[-1]["content"] if messages else prompt
        predictions = self.pipe(text, candidate_labels=self.wordlist, multi_label=True)
        scores = predictions['scores']

        for score in scores:
            if score > self.threshold:
                return {
                    "status": "error",
                    "validation_message": "Blocked Word Found",
                    "validation_passed": False,
                    "validated_output": messages if messages else prompt
                }
        return {
            "status": "success",
            "validation_passed": True,
            "validated_output": messages if messages else prompt
        }



from enum import Enum
import os


class ValidatorModules(Enum):
    WORD_FILTER = "word_filter"
    PROFANITY_FILTER = "profanity_filter"
    PII_FILTER = "pii_filter"
    PHRASE_FILTER = "phrase_filter"
    HARMFUL_CONTENT = "harmful_content"


VALIDATOR_MODULE_NAMES = [i.value for i in ValidatorModules]


class ConfigKeys(Enum):
    TYPE = 'type'
    SERVED_MODEL = "served_model"
    MODEL = "model"
    BLOCKED_CATEGORIES = "blocked_categories"
    CATEGORY_LEVELS = "category_levels"
    EXTRA = "extra"
    ENTITIES = "entities"
    LIB_NAME = "lib_name"
    WORDLIST = "wordlist"
    PHRASESLIST = "pharseslist"
    WORDLIST_FILE = "wordlist_file"
    WHITELIST_WORDS = "whitelist_words"
    OVERWRITE_WORDLIST = "overwrite_wordlist"
    PROMPT_TEMPLATE = "prompt_template"
    THRESHOLD = "threshold"
    NAME = "name"
    DESCRIPTION = "description"
    BLOCK_MESSAGE = "block_message"


class ValidatorTypes(Enum):
    EMBEDDINGS = "embedding"
    GUARD = "guard"
    LLM = "llm"
    ALGO = "algo"
    PRESIDIO = "presidio"
    LIB = "lib"
    CLASSIFIER = "classifier"


class LibTypes(Enum):
    BETTER_PROFANITY = "better-profanity"
    ALT_PROFANITY_CHECK = "alt-profanity-check"
    FUZZYSEARCH = "fuzzysearch"


class GenAIModels(Enum):
    E5_MISTRAL_7B_INSTRUCT = "e5-mistral-7b-instruct"
    LLAMA_GUARD_3_8B = "llama_guard_3_8b"
    LLAMA3_2_3B_INSTRUCT = "llama3_2_3b_instruct"


E2E_TIR_TEAM_ID = os.environ.get("E2E_TIR_TEAM_ID")
E2E_TIR_PROJECT_ID = os.environ.get("E2E_TIR_PROJECT_ID")
E2E_TIR_API_KEY = os.environ.get("E2E_TIR_API_KEY")
E2E_TIR_ACCESS_TOKEN = os.environ.get("E2E_TIR_ACCESS_TOKEN")

OPENAI_BASE_URL = f"https://infer.e2enetworks.net/project/p-{E2E_TIR_PROJECT_ID}" + "/genai/{model}/v1"
OPENAI_API_KEY = E2E_TIR_ACCESS_TOKEN




GUARD_CHAT_TEMP = \
    """
<|begin_of_text|><|start_header_id|>user<|end_header_id|>

<|image|> Task: Check if there is unsafe content in '{role}' messages in conversations according to our safety policy with the below categories.

<BEGIN UNSAFE CONTENT CATEGORIES>
{blocked_categories}
<END UNSAFE CONTENT CATEGORIES>

<BEGIN CONVERSATION>

{conversation}

<END CONVERSATION>

Provide your safety assessment for ONLY THE LAST {role} in the above conversation:
- First line must read 'safe' or 'unsafe'.
- If unsafe, a second line must include a comma-separated list of violated categories.<|eot_id|><|start_header_id|>assistant<|end_header_id|>
    """

GUARD_COMP_TEMP = \
    """
<|begin_of_text|><|start_header_id|>user<|end_header_id|>

<|image|> Task: Check if there is unsafe content in conversations according to our safety policy with the below categories.

<BEGIN UNSAFE CONTENT CATEGORIES>
{blocked_categories}
<END UNSAFE CONTENT CATEGORIES>

<BEGIN CONVERSATION>

{conversation}

<END CONVERSATION>

Provide your safety assessment for complete conversation given above:
- First line must read 'safe' or 'unsafe'.
- If unsafe, a second line must include a comma-separated list of violated categories.<|eot_id|><|start_header_id|>assistant<|end_header_id|>
    """



import re

from better_profanity import profanity
from openai import OpenAI
from profanity_check import predict

from tir_guardrails.constants import (ConfigKeys, ValidatorTypes, LibTypes, GenAIModels, OPENAI_BASE_URL, OPENAI_API_KEY)


class Validator:
    def __init__(self, config: dict):
        self.guard_type = ValidatorTypes(config.get(ConfigKeys.TYPE.value))
        self.extra: dict = config.get(ConfigKeys.EXTRA.value)

        if self.guard_type is ValidatorTypes.LIB:
            self.lib_name = LibTypes(self.extra.get(ConfigKeys.LIB_NAME.value))
            if self.lib_name == LibTypes.BETTER_PROFANITY:
                self.wordlist = self.extra.get(ConfigKeys.WORDLIST.value, None)
                self.wordlist_file = self.extra.get(ConfigKeys.WORDLIST_FILE.value, None)
                self.overwrite_wordlist = self.extra.get(ConfigKeys.OVERWRITE_WORDLIST.value, False)
                self.whitelist_words = self.extra.get(ConfigKeys.WHITELIST_WORDS.value, None)
                params = {}
                if self.whitelist_words:
                    params["whitelist_words"] = self.whitelist_words

                if self.wordlist_file:
                    if self.wordlist:
                        raise ValueError("Only one should be present either 'wordlist' or 'wordlist_file'")
                    profanity.load_censor_words_from_file(self.wordlist_file, **params)
                elif self.wordlist:
                    if self.overwrite_wordlist:
                        profanity.load_censor_words(self.wordlist, **params)
                    else:
                        profanity.load_censor_words(**params)
                        profanity.add_censor_words(self.wordlist)
                else:
                    profanity.load_censor_words(**params)

        elif self.guard_type is ValidatorTypes.LLM:
            self.served_model = GenAIModels(self.extra.get(ConfigKeys.SERVED_MODEL.value))
            self.prompt_template = self.extra.get(ConfigKeys.PROMPT_TEMPLATE.value)
            self.client = OpenAI(
                api_key=OPENAI_API_KEY,
                base_url=OPENAI_BASE_URL.format(model=self.served_model.value)
            )

    def validate(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")
        if self.guard_type == ValidatorTypes.LLM:
            return self.__validate_vllm(messages=messages, prompt=prompt)
        elif self.guard_type == ValidatorTypes.LIB:
            if self.lib_name == LibTypes.BETTER_PROFANITY:
                return self.__validate_better_profanity(messages=messages, prompt=prompt)
            elif self.lib_name == LibTypes.ALT_PROFANITY_CHECK:
                return self.__validate_alt_profanity_check(messages=messages, prompt=prompt)

    def __validate_vllm(self, messages=None, prompt=None):
        text = ""
        if messages:
            for i, msg in enumerate(messages):
                if i % 2 == 0:
                    text = text + "User: "
                else:
                    text = text + "Agent: "
                content = msg.get('content')
                if type(content) is str:
                    text = text + content + "\n"
                else:
                    raise NotImplementedError
        else:
            text = prompt
        prompt = self.prompt_template.format(text)
        try:
            completion = self.client.completions.create(
                model=self.served_model.value,
                prompt=prompt,
                max_tokens=3
            )
            model_response = completion.choices[0].text
            validation_passed = self.__parse_model_response(model_response)
            response = {
                "status": "success",
                "validation_passed": validation_passed,
                "validated_output": messages if messages else prompt
            }
            if not validation_passed:
                response["validation_message"] = "Conversation Contains Profane language"
            return response
        except Exception as e:
            # TODO Logging
            print(e)
            return {
                "status": "error",
                "validation_message": f"Following error occured {str(e)}",
                "validation_passed": False,
                "validated_output": messages if messages else prompt
            }

    def __validate_better_profanity(self, messages=None, prompt=None):
        text = messages[-1]["content"] if messages else prompt
        is_profane = profanity.contains_profanity(text)
        if is_profane:
            return {
                "status": "error",
                "validation_message": "Profane Language found",
                "validation_passed": False,
                "validated_output": messages if messages else prompt
            }
        return {
            "status": "success",
            "validation_passed": True,
            "validated_output": messages if messages else prompt
        }

    def __validate_alt_profanity_check(self, messages=None, prompt=None):
        text = messages[-1]["content"] if messages else prompt
        is_profane = int(predict([text])[0])
        if is_profane:
            return {
                "status": "error",
                "validation_message": "Profane Language found",
                "validation_passed": False,
                "validated_output": messages if messages else prompt
            }
        return {
            "status": "success",
            "validation_passed": True,
            "validated_output": messages if messages else prompt
        }

    def __parse_model_response(self, model_response):
        chunks = [item for item in model_response.replace("\n", " ").split(" ") if item]
        no_regex = r"(?i)no"
        check_no = [chunk for chunk in chunks if re.search(no_regex, chunk)]
        if len(check_no) >= 1:
            return True
        return False



import importlib

from tir_guardrails.constants import ConfigKeys, VALIDATOR_MODULE_NAMES


class Guard:
    def __init__(self, config: dict):
        self.block_message = config.pop(ConfigKeys.BLOCK_MESSAGE.value)
        self.validators: list = []
        module_name = "tir_guardrails.{}.validator"
        for validator_name in config.keys():
            if validator_name in VALIDATOR_MODULE_NAMES:
                module = importlib.import_module(module_name.format(validator_name))
                validator_class = getattr(module, "Validator")
                validator = validator_class(config.get(validator_name))
                self.validators.append(validator)

    def validate(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")
        validation_passed = True
        is_messages = True if messages else False

        latest_validated_input = messages if is_messages else prompt
        for validator in self.validators:
            kwargs = {"messages" if is_messages else "prompt": latest_validated_input}
            result = validator.validate(**kwargs)
            validation_passed = validation_passed and result["validation_passed"]
            latest_validated_input = result["validated_output"]

        if validation_passed:
            return {
                "status": "success",
                "validation_passed": validation_passed,
                "validated_output": latest_validated_input
            }

        return {
            "status": "error",
            "validation_passed": validation_passed,
            "validated_output": latest_validated_input
        }



from setuptools import find_packages, setup

setup(
    name="tir_guardrails",
    version="0.0.7.dev",
    description="TIR Guardrails is a safety framework for monitoring and mitigating harmful or inappropriate outputs in AI systems",
    author='Atharva Pakade',
    author_email='pakade310@gmail.com',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=[
        "better_profanity",
        "presidio_analyzer[transformers]",
        "presidio_anonymizer",
        "alt-profanity-check",
        "openai",
        "fuzzysearch"
    ],
    classifiers=[
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    license='Apache License 2.0'
)



import re

from openai import OpenAI

from tir_guardrails.constants import (ConfigKeys, ValidatorTypes, GenAIModels,
                                      OPENAI_BASE_URL, OPENAI_API_KEY)
from tir_guardrails.harmful_content.prompt_template import GUARD_CHAT_TEMP, GUARD_COMP_TEMP


class Validator:
    def __init__(self, config: dict):
        self.guard_type = ValidatorTypes(config.get(ConfigKeys.TYPE.value))
        self.served_model = GenAIModels(config.get(ConfigKeys.SERVED_MODEL.value))
        self.extra: dict = config.get(ConfigKeys.EXTRA.value)

        if ValidatorTypes.GUARD == self.guard_type:
            self.blocked_categories: list[dict] = self.extra.get(ConfigKeys.BLOCKED_CATEGORIES.value)
            self.client = OpenAI(
                api_key=OPENAI_API_KEY,
                base_url=OPENAI_BASE_URL.format(model=self.served_model.value)
            )
        elif ValidatorTypes.LLM == self.guard_type:
            raise NotImplementedError

    def validate(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")

        if ValidatorTypes.GUARD == self.guard_type:
            return self.__validate_guard(messages=messages, prompt=prompt)

    def get_prompt(self, messages=None, prompt=None):
        if (messages or prompt) is None:
            raise ValueError("Either 'messages' or 'prompt' must be provided.")

        if ValidatorTypes.GUARD == self.guard_type:
            category_pompt: str = "\n"
            for category in self.blocked_categories:
                category_pompt = category_pompt + category.get("name")
                des = category.get('description', None)
                if des:
                    category_pompt = category_pompt + f" {des}"
                category_pompt = category_pompt + "\n"

            conv_prompt: str = ""
            if messages:
                for i, msg in enumerate(messages):
                    if i % 2 == 0:
                        conv_prompt = conv_prompt + "User: "
                    else:
                        conv_prompt = conv_prompt + "Agent: "
                    content = msg.get('content')
                    if type(content) is str:
                        conv_prompt = conv_prompt + content
                    else:
                        raise NotImplementedError
                    conv_prompt = conv_prompt + "\n\n"

                return GUARD_CHAT_TEMP.format(
                    role="User" if (len(messages) % 2) == 1 else "Agent",
                    blocked_categories=category_pompt,
                    conversation=conv_prompt
                )
            elif prompt:
                conv_prompt = conv_prompt + prompt + "\n\n"
                return GUARD_COMP_TEMP.format(
                    blocked_categories=category_pompt,
                    conversation=conv_prompt
                )

    def __validate_guard(self, messages=None, prompt=None):
        guard_prompt = self.get_prompt(messages=messages, prompt=prompt)
        if self.guard_type == ValidatorTypes.GUARD:
            try:
                completion = self.client.completions.create(
                    model=self.served_model.value,
                    prompt=guard_prompt
                )
                model_response = completion.choices[0].text
                validation_passed, categories_found = self.__parse_model_response(model_response)
                response = {
                    "status": "success",
                    "validation_passed": validation_passed,
                    "validated_output": prompt if prompt else messages
                }
                if not validation_passed:
                    response["validation_message"] = "Following unsafe content categories found " + ",".join(categories_found)
                return response
            except Exception as e:
                # TODO Logging
                print(e)
                return {
                    "status": "error",
                    "validation_message": f"Following error occured {str(e)}",
                    "validation_passed": False,
                    "validated_output": prompt if prompt else messages
                }

    def __parse_model_response(self, model_response: str):
        chunks = [item for item in model_response.strip(' ').split("\n") if item]
        regex = r"^S([1-9]|[1-4][0-9])$"
        cat_keys_found = [chunk for chunk in chunks if re.search(regex, chunk)]
        is_safe = True if chunks[0] == "safe" else False
        categories_found = []
        for category in self.blocked_categories:
            for cat_key_found in cat_keys_found:
                if category.get("name").startswith(f"{cat_key_found}:"):
                    categories_found.append(category.get("name"))
        return is_safe, categories_found
